@@ -187,7 +187,7 @@ class P4Client(object):
         marked for delete
         :return: *list* P4Files
         """
-        file_list = [file_list] if not isinstance(file_list, list) else file_list
+        file_list = convert_to_list(file_list) if not isinstance(file_list, list) else file_list
 
         if self.host_online():
             fstat_output = self.run_cmd2("fstat", file_list)
@@ -281,6 +281,8 @@ class P4Client(object):
         :param changelist: *string* or *int* changelist description or changelist number
         :return: *list* of info dictionaries
         """
+        file_list = convert_to_list(file_list) if not isinstance(file_list, list) else file_list
+
         changelist = self.__ensure_changelist(changelist)
 
         info_dicts = self.run_cmd2("reopen", ["-c", changelist] + file_list)
@@ -314,18 +316,22 @@ class P4Client(object):
         :param file_list: *list* files
         :return: *list* of info dictionaries
         """
+        file_list = convert_to_list(file_list) if not isinstance(file_list, list) else file_list
+        if not self.silent: self.__validate_file_list(file_list)
         info_dicts = self.run_cmd2("revert", file_list)
         return info_dicts
 
-    def revert_folders(self, folder_list=[]):
+    def revert_folders(self, folder_list):
         """
         Recursively reverts complete folders
 
         :param folder_list: *list* folder paths
         :return: *list* of info dicts
         """
-        cleaned_folder_list = []
+        folder_list = convert_to_list(folder_list) if not isinstance(folder_list, list) else folder_list
+        if not self.silent: self.__validate_file_list(folder_list)
 
+        cleaned_folder_list = []
         for folder in folder_list:
             folder = folder.replace("\\", "/")
             folder = folder.rstrip("/")
@@ -335,15 +341,25 @@ class P4Client(object):
         info_dicts = self.run_cmd2("revert", [] + cleaned_folder_list)
         return info_dicts
 
-    def sync_folders(self, folder_list=[]):
+    def revert_changelist(self, changelist="default"):
+        """
+        Reverts all files in a given changelist
+        :param changelist: string or int value
+        """
+        files = self.get_files_in_changelist(changelist)
+        self.revert_files(files)
+
+    def sync_folders(self, folder_list):
         """
         Recursively syncs complete folders
 
         :param folder_list: *list* folder paths
         :return: *list* of info dicts
         """
-        cleaned_folder_list = []
+        folder_list = convert_to_list(folder_list) if not isinstance(folder_list, list) else folder_list
+        if not self.silent: self.__validate_file_list(folder_list)
 
+        cleaned_folder_list = []
         for folder in folder_list:
             folder = folder.replace("\\", "/")
             folder = folder.rstrip("/")
@@ -353,16 +369,21 @@ class P4Client(object):
         info_dicts = self.run_cmd2("sync", [] + cleaned_folder_list)
         return info_dicts
 
-    def sync_files(self, file_list=[], verify=True):
+    def sync_files(self, file_list, verify=True, force=False):
         """
         Syncs files
 
         :param file_list: *list*
         :param verify: *bool* if true, checks that file synced files exist on disk. Throws a warning if they don't
         This could happen when a synced file is deleted locally
+        :param force: *bool* force sync
         :return: *list* of info dicts
         """
-        info_dicts = self.run_cmd2("sync", [] + file_list)
+        file_list = convert_to_list(file_list) if not isinstance(file_list, list) else file_list
+        initial_arg_list = ["-f"] if force else []
+        if not self.silent: self.__validate_file_list(file_list)
+
+        info_dicts = self.run_cmd2("sync", initial_arg_list + file_list)
 
         if verify:
             local_file_paths = self.get_local_paths(file_list)
@@ -373,7 +394,7 @@ class P4Client(object):
 
         return info_dicts
 
-    def delete_files(self, file_list=[], changelist="default"):
+    def delete_files(self, file_list, changelist="default"):
         """
         Marks files for delete
 
@@ -381,9 +402,22 @@ class P4Client(object):
         :param changelist: *string* or *int* changelist number
         :return: *list* of info dictionaries
         """
+        file_list = convert_to_list(file_list) if not isinstance(file_list, list) else file_list
+        if not self.silent: self.__validate_file_list(file_list)
         info_dicts = self.run_cmd2("delete", ["-c", changelist] + file_list)
         return info_dicts
 
+    def delete_changelist(self, changelist="default", perfect_match_only=False, case_sensitive=False):
+        """
+        Deletes a changelist via cl num or description
+        :param changelist:
+        :return:
+        """
+        cl_num = self.get_pending_changelists(changelist, perfect_match_only, case_sensitive)
+        for cl in cl_num:
+            info_dicts = self.run_cmd2('change', ['-d', cl])
+            # TODO: Break down info dicts and look for errors
+
     def get_shelved_files(self):
         """
         Returns a list that holds other lists where index 0 is the depot path of the shelved file and index 1
@@ -403,15 +437,16 @@ class P4Client(object):
 
         return files_and_cl
 
-    def add_or_edit_files(self, file_list=[], changelist="default"):
+    def add_or_edit_files(self, file_list, changelist="default"):
         """
         Marks the files in file_list for add if they are new, or edit if they are already versioned
 
-        :param file_list: *list*
+        :param file_list: *list* (ideally)
         :param changelist: *string* or *int* changelist number or description. Will be made if it doesn't exist.*string* or *int* changelist number
         :return: *list* of info dictionaries
         """
-        changelist = self.__ensure_changelist(changelist)
+        file_list = convert_to_list(file_list) if not isinstance(file_list, list) else file_list
+        if not self.silent: self.__validate_file_list(file_list)
 
         files_for_add = []
         files_for_checkout = []
@@ -419,6 +454,8 @@ class P4Client(object):
 
         p4files = self.files_to_p4files(file_list, allow_invalid_files=True)
         for p4file in p4files:
+            if p4file.is_checked_out():
+                continue
             if p4file.is_local_only():
                 files_for_add.append(p4file.get_local_file_path())
             elif p4file.get_depot_file_path() is not None:
@@ -447,7 +484,7 @@ class P4Client(object):
                 return int(self.__get_dict_value(info_dict, "change"))
         return -1
 
-    def edit_files(self, file_list=[], changelist="default"):
+    def edit_files(self, file_list, changelist="default"):
         """
         Marks the files in file_list for edit
 
@@ -455,6 +492,9 @@ class P4Client(object):
         :param changelist: *string* or *int* changelist number or description. Will be made if it doesn't exist.
         :return: *list* of info dictionaries
         """
+        file_list = convert_to_list(file_list) if not isinstance(file_list, list) else file_list
+        if not self.silent: self.__validate_file_list(file_list)
+
         changelist = self.__ensure_changelist(changelist)
 
         info_dicts = self.run_cmd2("edit", ["-c", changelist] + file_list)
@@ -463,13 +503,16 @@ class P4Client(object):
                 print(self.__get_dict_value(info_dict, "data"))
         return info_dicts
 
-    def add_files(self, file_list=[], changelist="default"):
+    def add_files(self, file_list, changelist="default"):
         """
         Marks the files in file_list for add
         :param file_list: *list*
         :param changelist:  *string* or *int* changelist number or description. Will be made if it doesn't exist.
         :return: *list* of info dictionaries
         """
+        file_list = convert_to_list(file_list) if not isinstance(file_list, list) else file_list
+        if not self.silent: self.__validate_file_list(file_list)
+
         changelist = self.__ensure_changelist(changelist)
 
         info_dicts = self.run_cmd2("add", ["-c", changelist] + file_list)
@@ -507,7 +550,7 @@ class P4Client(object):
         :param case_sensitive: *bool*
         :return: *list* with changelist numbers as ints
         """
-        info_dicts = self.run_cmd2("changes", ["-s", "pending", "-u", self.user, "-c", self.client])
+        info_dicts = self.run_cmd2("changes", ["-l", "-s", "pending", "-u", self.user, "-c", self.client])
         changelists = []
 
         for info_dict in info_dicts:
@@ -596,10 +639,36 @@ class P4Client(object):
         :param paths: *list* of file paths
         :return: *list* of depot paths
         """
-        info_dicts = self.run_cmd2("where", paths)
+
+        # strip out revision specifier from paths
+        no_rev_paths = []
+        for path in paths:
+            path_without_ext, path_ext = os.path.splitext(path)
+            if "#" in path_ext:
+                path_ext = path_ext.split("#")[0]
+            no_rev_paths.append("{}{}".format(path_without_ext, path_ext))
+
+        info_dicts = self.run_cmd2("where", no_rev_paths)
         local_paths = [self.__get_dict_value(info, "path") for info in info_dicts]
         return local_paths
 
+    def get_history(self, paths):
+        """
+        Returns changes for the given paths
+
+        :param paths: *list* of file paths
+        :return: *list* of change info
+        """
+        info_dicts = []
+        for path in paths:
+            info_dicts.extend(self.run_cmd2("changes", ["-l", path]))
+
+        # decode from bytes
+        if sys.version_info.major > 2:
+            info_dicts = decode_dictionaries(info_dicts)
+
+        return info_dicts
+
     def host_online(self):
         """
         Checks if the host for this client is online
@@ -726,6 +795,21 @@ class P4Client(object):
             current_dir = os.path.abspath(current_dir + os.path.sep + os.pardir)
         return False
 
+    def __validate_file_list(self, file_list):
+        """
+        Validation function to ensure correct files are being synced to correct workspaces & clients etc.
+        Is an extendable function
+        :param file_list: List of files to iterate
+        :return:
+        """
+        file_list = convert_to_list(file_list) if not isinstance(file_list, list) else file_list
+
+        # Easy utility to check that the file is underneath the correct perforce root
+        # Quicker than waiting for the result of a p4 fstat
+        for f in file_list:
+            if not f.lower().startswith(self.perforce_root.lower()):
+                raise Exception(f'{f} is not under perforce root: {self.perforce_root}')
+
 
 def split_list_into_strings_of_length(input_list, max_length=100):
     """
@@ -758,6 +842,25 @@ def split_list_into_strings_of_length(input_list, max_length=100):
     return clamped_str_list
 
 
+def decode_dictionaries(info_dicts):
+    """
+    Decode list of dictionary keys and values into unicode from bytes
+    """
+    decoded_dicts = []
+    for info_dict in info_dicts:
+        decoded_dict = {k.decode(): v.decode() for k, v in info_dict.items()}
+        decoded_dicts.append(decoded_dict)
+    return decoded_dicts
+
+
+def convert_to_list(value):
+    if isinstance(value, tuple):
+        converted = [v for v in value]
+    else:
+        converted = [value]
+    return converted
+
+
 class HostOnline(object):
     def __init__(self, host, port):
         self.host = host
